import re
from collections import OrderedDict

from django import forms
from django.utils.translation import ugettext_lazy as _

import phonenumbers
from django_select2.forms import ModelSelect2MultipleWidget, ModelSelect2Widget
from django_starfield import Stars
from material import Fieldset, Layout, Row

from aleksis.core.forms import AccountRegisterForm
from aleksis.core.mixins import ExtensibleForm
from aleksis.core.models import Group, Person
from aleksis.core.util.core_helpers import get_site_preferences

from .models import Event, EventRegistration, FeedbackAspect, Voucher

COMMENT_CHOICES = [
    ("first", _("Only first name")),
    ("first_age", _("First name and age")),
    ("first_last_age", _("First name, last name and age")),
]

TEMPLATE_CHOICES = [
    ("list_sign", _("Signature list")),
    ("list_participants", _("Participants list")),
    ("corona", _("Corona attendance list")),
]

LICENCE_CHOICES = [
    ("CC-BY-4.0+", _("Creative Commons with attribution, 4.0 or later")),
    (
        "CC-BY-SA-4.0+",
        _(
            "Creative Commons with attribution and distribution only"
            "under the same conditions, 4.0 or later"
        ),
    ),
]

NEWSLETTER_CHOICES = get_site_preferences()["paweljong__newsletter_choices"].split(",")


def is_phonenumber(number):
    try:
        phonenumbers.parse(number, "DE")
    except BaseException:
        raise forms.ValidationError(_("%s is not a valid phone number") % number)


def is_valid_voucher_for(event_cn):
    def _is_valid_voucher(code):
        if Voucher.objects.filter(code=code, event_cn=event_cn, used=False).count() == 0:
            raise forms.ValidationError(_("The voucher is invalid!"))

    return _is_valid_voucher


def clean_phonenumber(field):
    def _clean_phonenumber(obj):
        value = obj.cleaned_data[field]

        if value:
            pn = phonenumbers.parse(value, "DE")
            value = phonenumbers.format_number(pn, phonenumbers.PhoneNumberFormat.E164)

        return value

    return _clean_phonenumber


class EventAdditionalSurveyForm(forms.Form):
    def __init__(self, event, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_fields_from_ldap(event)

    def add_fields_from_ldap(self, event, before=None):
        new_fields = OrderedDict()

        for field in event.registration_fields:
            label, help_text, *choices = field.split("|")
            var = re.sub(r"[^A-Za-z0-9]|^(?=\d)", "_", label)

            if choices:
                choices_map = [
                    (re.sub(r"[^A-Za-z0-9]|^(?=\d)", "_", choice), choice) for choice in choices
                ]
                field_attr = forms.ChoiceField(
                    label=label,
                    help_text=help_text,
                    choices=choices_map,
                    required=False,
                )
            else:
                field_attr = forms.CharField(label=label, help_text=help_text, required=False)

            new_fields[var] = field_attr

        if before:
            before_field_index = list(self.fields.keys()).index(before)
            field_names = list(self.fields.keys())
            new_field_order = (
                field_names[: before_field_index - 1]
                + list(new_fields.keys())
                + field_names[before_field_index - 1 :]
            )

        self.fields.update(new_fields)

        if before:
            self.order_fields(new_field_order)


class EventFeedbackForm(ExtensibleForm):
    class Meta:
        model = FeedbackAspect
        fields = []

    layout = Layout(
        Fieldset(
            _("Comments"),
            Row("comment_private", "comment_public", "comment_public_info"),
        ),
        Fieldset(
            _("Photos"),
            Row("photos", "photos_licence"),
        ),
        Fieldset(
            _("Feedback aspects"),
        ),
    )

    comment_private = forms.CharField(
        required=False,
        label=_("Comment for the team"),
        help_text=_(
            "This comment is for the team only. You can write down everything you"
            "would like to give us as feedback here."
        ),
        widget=forms.Textarea,
    )

    comment_public = forms.CharField(
        required=False,
        label=_("Comment for the website"),
        help_text=_(
            "This comment is for the report on our website. Tell in detail about what"
            "you experienced, what you liked, what you learned and everything else"
            "you can think of."
        ),
        widget=forms.Textarea,
    )

    comment_public_info = forms.ChoiceField(
        label=_("Information in the comment"),
        choices=COMMENT_CHOICES,
        help_text=_("What information would you like to use to publish your comment?"),
    )

    photos = forms.ImageField(
        label=_("Photos"),
        widget=forms.ClearableFileInput(attrs={"multiple": True}),
        required=False,
        help_text=_(
            "If you want to contribute photos to the report, you can upload them here. You can"
            "select multiple files in most file selection dialogs with CTRL + click."
        ),
    )

    photos_licence = forms.ChoiceField(
        label=_("Photo licence"),
        choices=LICENCE_CHOICES,
        required=False,
        help_text=_("If you upload photos, choose a license here."),
    )

    def __init__(self, event, *args, **kwargs):
        super(EventFeedbackForm, self).__init__(*args, **kwargs)
        self._event = event

        for aspect in event.feedback_aspects.all():
            field = forms.IntegerField(
                widget=Stars,
                required=False,
            )
            self.fields[aspect.aspect] = field
            node = Fieldset(f"{aspect.aspect}", f"{aspect.aspect}")
            self.add_node_to_layout(node)


class EditEventForm(forms.ModelForm):
    """Form to create or edit an event."""

    layout = Layout(
        Fieldset(
            _("Base data"),
            "group",
            Row("display_name", "description"),
            Row("place", "published"),
            Fieldset(_("Date data"), Row("date_event", "date_registration", "date_retraction")),
            Fieldset(_("Event details"), Row("cost", "max_participants")),
            Fieldset(_("Feedback aspects"), "feedback_aspects"),
        ),
    )

    class Meta:
        model = Event
        exclude = []
        widgets = {
            "group": ModelSelect2Widget(
                search_fields=["name__icontains", "short_name__icontains"],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
            "feedback_aspects": ModelSelect2MultipleWidget(
                search_fields=["aspect__icontains"],
                attrs={"data-minimum-input-length": 0, "class": "browser-default"},
            ),
        }


class EditVoucherForm(forms.ModelForm):
    """Form to edit and create vouchers."""

    class Meta:
        model = Voucher
        exclude = ["code", "used_person_uid", "used", "deleted"]
        help_texts = {
            "event": _("Event the voucher is valid for."),
            "person": _("Person the voucher is valid for."),
            "discount": _("Voucher discount."),
        }


class GenerateListForm(forms.Form):
    """Form to create a list of participants of a group."""

    group = forms.ModelChoiceField(
        label=_("Group"),
        queryset=Group.objects.all(),
        help_text=_("Select group to generate list."),
    )

    template = forms.ChoiceField(
        label=_("Template"),
        choices=TEMPLATE_CHOICES,
        help_text=_("Select template to generate list."),
    )

    landscape = forms.BooleanField(
        label=_("Landscape"),
        help_text=_("Select if output should be in landscape."),
    )


class RegisterEventForm(forms.ModelForm):
    """Form to register for an event."""

    layout = Layout(
        Fieldset(
            _("Address data"),
            Row("street", "housenumber"),
            Row("postal_code", "place"),
        ),
        Fieldset(
            _("Contact details"),
            Row("mobile_number", "email"),
        ),
        Fieldset(
            _("Personal data"),
            Row("date_of_birth", "sex"),
        ),
        Fieldset(
            _("School details"),
            Row("school", "school_place", "school_class"),
        ),
        Fieldset(
            _("Guardians personal data"),
            Row("guardian_first_name", "guardian_last_name"),
        ),
        Fieldset(
            _("Guardians contact details"),
            Row("guardian_email", "guardian_mobile_number"),
        ),
        Fieldset(
            _("General event information"),
            Row("event", "person"),
            Row("comment", "channel"),
        ),
        Fieldset(
            _("Financial data"),
            "voucher_code",
            Row("iban", "donation", "accept_sepa"),
        ),
        Fieldset(
            _("Declaration of consent"),
            Row("accept_terms", "accept_data", "accept_general_terms"),
        ),
    )

    class Meta:
        model = EventRegistration
        exclude = ["date_registred", "voucher"]
        help_texts = {
            "voucher": _(
                "If you have a voucher for the event, enter the code here."
                "It will be charged automatically."
            ),
            "donation": (
                "Our association would like to offer all children and young"
                "people the opportunity to participate in our events. Often,"
                "however, the family fee cannot be paid. We therefore have a"
                "budget from which we can promote participation after we have"
                "carefully examined the necessity and eligibility. We rely on"
                "donations for this budget. If you would like to donate a voluntary"
                "additional amount for this budget, please indicate this here. We do"
                "not permanently save whether and if so in what amount donations are"
                "made and also not within the association, e.g. passed on to leisure supervisors."
            ),
            "accept_sepa": _(
                "Parents: I authorize the creditor  e.V., Rochusstr. 2-4, 53123 Bonn with"
                "creditor ID DE70FZT00001497650, to collect the participant fee from my account"
                "once using the SEPA core direct debit. At the same time, I instruct my bank to"
                "redeem the SEPA core direct debit withdrawn from my account by  e.V."
            ),
            "iban": _(
                "If your parents want to pay by SEPA direct debit,"
                "please let them fill out this field."
            ),
            "accept_terms": _(
                "Parents: My child filled out the registration form together with me, but myself,"
                "and I agree to the participation, the terms of use and the terms and conditions."
                "I am aware that the registration is binding and that withdrawal is only possible"
                "in exceptional cases with a valid reason. In addition, I agree to pay the"
                "participation fee in advance and agree to the reimbursement guidelines"
                "mentioned above."
            ),
            "accept_data": _(
                "I consent to the processing of my data as stated in the"
                "terms of use and all the data provided is correct. If I am under the"
                "age of 16, my parents also agree to this and I can prove this on request"
                "(e.g. by making contact with my parents)."
            ),
            "accept_general_terms": _("I agree with the" "AGB and have read them."),
            "channel": _("How did you find out about the event?"),
        }

    guardian_first_name = forms.CharField(
        label=_("Guardians first name"),
        help_text=_(
            "Please enter the first name of the legal guardian who will fill in the registration"
            "with you and who can be reached during the event in an emergency."
        ),
    )

    guardian_last_name = forms.CharField(
        label=_("Guardians last name"),
        help_text=_(
            "Please enter the last name of the legal guardian who will fill in the registration"
            "with you and who can be reached during the event in an emergency."
        ),
    )

    guardian_mobile_number = forms.CharField(
        label=_("Guardians mobile number"),
        help_text=_(
            "We need the mobile phone number for emergencies if we"
            "urgently need to reach your parents during the event."
        ),
    )

    guardian_email = forms.EmailField(
        label=_("Guardians email address"),
    )

    voucher_code = forms.CharField(
        label=_("Voucher code"),
        help_text=_("If you have a voucher code, type it in here."),
        required=False,
    )

    street = forms.CharField(
        label=_("Street"),
    )

    housenumber = forms.CharField(
        label=_("Housenumber"),
    )

    postal_code = forms.CharField(
        label=_("Postal code"),
    )

    place = forms.CharField(
        label=_("Place"),
    )

    mobile_number = forms.CharField(
        label=_("Mobile number"),
        required=False,
        help_text=_(
            "Your mobile number helps us to reach you in an emergency during the event, e.g."
            "if you are alone with your group at a conference or similar. If you don't have a"
            "cell phone, you can leave the field blank."
        ),
        validators=[is_phonenumber],
    )

    date_of_birth = forms.DateField(
        label=_("Date of birth"),
    )

    sex = forms.ChoiceField(
        label=_("Sex"),
        help_text=_(
            "For various reasons, e.g. because we have to keep gender segregation during the night"
            "for legal reasons, we need to know if you are a boy or a girl. With some names this is"
            "not always immediately recognizable, so we ask you to indicate it here."
        ),
        choices=Person.SEX_CHOICES,
        initial=None,
    )

    email = forms.EmailField(
        label=_("Email address"),
    )

    school = forms.CharField(
        label=_("School"),
        help_text=_("Please enter the name of your school as exactly as it should be written."),
    )

    school_place = forms.CharField(
        label=_("School place"),
        help_text=_("Enter the place (city) where your school is located (without a district)."),
    )

    school_class = forms.CharField(
        label=_("School class"),
        help_text=_("Please enter the class you are going to (e.g. 8a)."),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields["event"].disabled = True
        self.fields["person"].disabled = True
        self.fields["accept_terms"].required = True
        self.fields["accept_general_terms"].required = True
        self.fields["accept_data"].required = True


class EditEventRegistrationForm(forms.ModelForm):

    layout = Layout(
        Fieldset(
            _("General event information"),
            Row("event", "person"),
            Row("comment", "channel"),
        ),
        Fieldset(
            _("Financial data"),
            "voucher_code",
            Row("iban", "donation", "accept_sepa"),
        ),
        Fieldset(
            _("Declaration of consent"),
            Row("accept_terms", "accept_data", "accept_general_terms"),
        ),
    )

    class Meta:
        model = EventRegistration
        help_texts = {
            "voucher": _(
                "If you have a voucher for the event, enter the code here."
                "It will be charged automatically."
            ),
            "donation": (
                "Our association would like to offer all children and young"
                "people the opportunity to participate in our events. Often,"
                "however, the family fee cannot be paid. We therefore have a"
                "budget from which we can promote participation after we have"
                "carefully examined the necessity and eligibility. We rely on"
                "donations for this budget. If you would like to donate a voluntary"
                "additional amount for this budget, please indicate this here. We do not"
                "permanently save whether and if so in what amount donations are made"
                "and also not within the association, e.g. passed on to leisure supervisors."
            ),
            "accept_sepa": _(
                "Parents: I authorize the creditor  e.V., Rochusstr. 2-4, 53123 Bonn with"
                "creditor ID DE70FZT00001497650, to collect the participant fee from my account"
                "once using the SEPA core direct debit. At the same time, I instruct my bank"
                "to redeem the SEPA core direct debit withdrawn from my account by  e.V."
            ),
            "iban": _(
                "If your parents want to pay by SEPA direct debit,"
                "please let them fill out this field."
            ),
            "accept_terms": _(
                "Parents: My child filled out the registration form together with me, but myself,"
                "and I agree to the participation, the terms of use and the terms and conditions."
                "I am aware that the registration is binding and that withdrawal is only possible"
                "in exceptional cases with a valid reason. In addition, I agree to pay the"
                "participation fee in advance and agree to the reimbursement"
                "guidelines mentioned above."
            ),
            "accept_data": _(
                "I consent to the processing of my data as stated in the"
                "terms of use and all the data provided is correct. If I am under"
                "the age of 16, my parents also agree to this and I can prove this on"
                "request (e.g. by making contact with my parents)."
            ),
            "accept_general_terms": _("I agree with the" "AGB and have read them."),
            "channel": _("How did you find out about the event?"),
        }
        exclude = []


class EditFeedbackAspectForm(forms.ModelForm):
    class Meta:
        model = FeedbackAspect
        exclude = []


date_of_birth = forms.DateField(label=_("Date of birth"))
extend_register_form = Fieldset(date_of_birth)
AccountRegisterForm.add_node_to_layout(extend_register_form)
